



test.pyutilib --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
pyutilib_test_driver --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
command -v lbin
IF %ERRORLEVEL% NEQ 0 exit /B 1
command -v lpython
IF %ERRORLEVEL% NEQ 0 exit /B 1
dispatch_srvr --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
